use aphid::{imbalance, BranchLength, Filters};

use super::{
    justify, ArrayVec, Colorize, GeneTree, Interner, ResolvedSymbol, SectionAnalysis, SectionLca,
    ToString, TopologicalAnalysis, TreeSymbol,
};

//==================================================================================================
// Display tree analysis results for the first, topological pass,
// one line per "filter" (loosely defined yet).
// Return false for every filter not passed.
// This function is essentially a zoom back into the filtering process.
// Con 1: it duplicates most of the logic there.
// Pro 1: this is a good opportunity to keep the filtering procedure *modular*
//        so that this fine-grained breakdown remains possible,
//        which in turns improves future flexibility.
// Pro 2: this is not strictly necessary for the procedure to run,
//        so it can just be turned off if performances eventually get in the way.
#[allow(clippy::too_many_lines)] // Ongoing experiment.
pub(super) fn display_topological_tree_analysis(
    id: TreeSymbol,
    raw: &GeneTree,
    pruned: &GeneTree,
    filters: &Filters,
    analysis: &TopologicalAnalysis,
    included: bool,
    interner: &Interner,
) -> (bool, bool, Option<bool>) {
    use SectionAnalysis as S;
    let header = format!("{} nodes, pruned to {}.", raw.len(), pruned.len());
    let resolve = |vec: &[_]| {
        vec.iter()
            .map(|&sp| ResolvedSymbol::new(sp, interner))
            .collect::<Vec<_>>()
    };

    //----------------------------------------------------------------------------------------------
    // Retrieve (ok, message) status for every "filter" (loosely defined).

    // Triplet filter.
    let triplet = match &analysis.triplet {
        S::AllMissing => (false, "None found.".into()),
        S::MissingSome(missing, _) => (false, format!("Missing species: {:?}.", resolve(missing))),
        S::AllFound(SectionLca { id, paraphyletic }) => {
            if paraphyletic.is_empty() {
                (true, format!("LCA found: {id}."))
            } else {
                let names = resolve(paraphyletic);
                (false, format!("Paraphyletic with {names:?}."))
            }
        }
    };

    // Outgroup filter.
    let mut outgroup = {
        let mut lines = ArrayVec::<_, 2>::new();
        match &analysis.outgroup {
            S::AllMissing => lines.push((false, "None found.".into())),
            S::MissingSome(missing, _) => {
                lines.push((true, format!("Missing species: {:?}.", resolve(missing))));
            }
            S::AllFound(_) => {}
        }
        match &analysis.outgroup {
            S::AllMissing => {}
            S::MissingSome(_, lca) | S::AllFound(lca) => {
                let SectionLca { id, paraphyletic } = lca;
                lines.push(if paraphyletic.is_empty() {
                    (true, format!("LCA found: {id}."))
                } else {
                    let names = resolve(paraphyletic);
                    (false, format!("Paraphyletic with {names:?}."))
                });
            }
        }
        lines
    };
    outgroup.sort_unstable_by_key(|o| o.0); // Error first, if any.

    // Treetop filter.
    let mut top = analysis.top.as_ref().map(|top| {
        if top.external.is_empty() {
            if let Some(internals) = &top.internal {
                if filters.triplet_other_monophyly && !internals.outgroup_side().is_empty() {
                    let names = resolve(internals.outgroup_side());
                    (
                        false,
                        format!("LCA(triplet, other) paraphyletic with: {names:?}"),
                    )
                } else {
                    let names = resolve(&internals.species);
                    (true, format!("Other species found: {names:?}"))
                }
            } else {
                (true, "No other species remain.".into())
            }
        } else {
            let names = resolve(&top.external);
            (false, format!("External species found: {names:?}."))
        }
    });

    //----------------------------------------------------------------------------------------------
    // Display.

    // Header line.
    let check = if included { "✔".green() } else { "🗙".red() };
    let name = interner.resolve(id).unwrap();
    eprintln!("{check} {}: {header}", name.blue());
    let indent = "  ";
    let tab = |p: &str| justify(p, 12, " ", '>');

    // Triplet line.
    let (ok, mut message) = triplet;
    let mut prefix = String::from("triplet:");
    if !ok {
        prefix = prefix.red().string();
        message = message.italic().string();
    }
    eprintln!("{indent}{} {message}", tab(&prefix));

    // Outgroup lines.
    let mut prefix = String::from("outgroup:");
    if !outgroup.iter().all(|o| o.0) {
        prefix = prefix.red().string();
    }
    eprint!("{indent}{}", tab(&prefix));
    for &(ok, ref message) in &outgroup {
        if ok {
            eprint!(" {message}");
        } else {
            eprint!(" {}", message.italic());
        }
    }
    eprintln!();

    // Treetop line.
    if let Some(t) = &mut top {
        let &mut (ref ok, ref mut message) = t;
        let mut a = String::from("treetop:");
        if !ok {
            a = a.red().string();
            *message = message.italic().string();
        }
        eprintln!("{indent}{} {message}", tab(&a));
    }

    (triplet.0, outgroup.iter().all(|o| o.0), top.map(|t| t.0))
}

//==================================================================================================
// Same, for the second, geometrical pass.
pub(super) fn display_geometrical_tree_analysis(
    id: TreeSymbol,
    gtree: &GeneTree,
    (mean_triplet, mean_rest): (BranchLength, BranchLength),
    global_shape: f64,
    max_clock_ratio: f64,
    included: bool,
    interner: &Interner,
) {
    // Loosely fit to previous formatting, but there is only one filter for now.
    // TODO: factorize formatting out if it grows.

    let clock_ratio = {
        let (triplet_longer, imbalance) = imbalance(mean_triplet, mean_rest);
        let (t, r) = ("'triplet' section", "'outgroup' and 'other' section");
        let (small, large) = if triplet_longer { (t, r) } else { (r, t) };
        let (pass, sratio) =
            gtree.is_shape_included(mean_triplet, mean_rest, global_shape, max_clock_ratio);
        if pass {
            (
                true,
                format!(
                    "Branch lengths in {large} are only {imbalance:.3} times longer than in {small}: \
                     shape ratio = {sratio:.3} < {max_clock_ratio:.3}."
                ),
            )
        } else {
            (
                false,
                format!(
                    "Branch lengths in {large} are {imbalance:.3} times longer than in {small}: \
                     shape ratio = {sratio:.3} ⩾ {max_clock_ratio:.3}."
                ),
            )
        }
    };

    // Header line.
    let check = if included { "✔".green() } else { "🗙".red() };
    let name = interner.resolve(id).unwrap();
    eprintln!("{check} {}:", name.blue());
    let indent = "  ";
    let tab = |p: &str| justify(p, 15, " ", '>');

    // Clock ratio line.
    let (ok, mut message) = clock_ratio;
    let mut prefix = String::from("clock ratio:");
    if !ok {
        prefix = prefix.red().string();
        message = message.italic().string();
    }
    eprintln!("{indent}{} {message}", tab(&prefix));
}
