// Main entry point into the CLI.

use std::{cmp::max, collections::HashSet, fmt::Write, path::PathBuf, process};

use aphid::{
    extract_local_triplet, imbalance,
    interner::{Interner, ResolvedSymbol, TreeSymbol},
    it_mean::SummedMean,
    ln_likelihood, optimize_likelihood,
    topological_analysis::{SectionAnalysis, SectionLca},
    Config, Error as AphidError, GeneTree, GeneTriplet, GenesForest, TopologicalAnalysis, VERSION,
};
use arrayvec::ArrayVec;
use clap::Parser;
use colored::{ColoredString, Colorize};
use lexical_sort::natural_lexical_cmp;

mod display_tree_analysis;
mod justify;
use display_tree_analysis::display_topological_tree_analysis;
use justify::{justify, terminal_width};

use crate::display_tree_analysis::display_geometrical_tree_analysis;

/// Aphid: distinguishing gene flow from incomplete lineage sorting.
#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// Path to the config file.
    #[arg(short, long)]
    config: PathBuf,
}

fn main() {
    eprintln!("Running Aphid v{VERSION}.");
    match run() {
        Ok(()) => {
            eprintln!("Success. {}", "✓".bold().green());
        }
        Err(e) => {
            eprintln!("{} {e}", "🗙".bold().red());
            process::exit(e.code());
        }
    }
}

#[allow(clippy::too_many_lines)] // TODO: factorize later: ongoing experiment.
fn run() -> Result<(), AphidError> {
    use AphidError as E;
    // Parse command line arguments.
    let args = Args::parse();

    let separator = || eprintln!("\n{:=<80}", "");

    eprintln!("Prepare string interner.");
    let mut interner = Interner::new();
    let interner = &mut interner;

    //==============================================================================
    // Read input
    let path = &args.config;
    eprintln!("Read config from {}.", format!("{path:?}").blue());
    let config = Config::from_file(path, interner)?;

    let path = &config.trees;
    eprintln!("Read gene trees from {}.", format!("{path:?}").blue());
    let forest = GenesForest::from_file(path, interner)?;

    //==============================================================================
    // Check input consistency.

    eprintln!("  Check config consistency.");

    // Check that all relevant species are actually found in the forest,
    // or they may be mispelled.
    let relevant_species = config.relevant_species().collect::<HashSet<_>>();
    let mut not_found = relevant_species.clone();
    for gtree in forest.trees() {
        for sp in gtree.species_symbols() {
            not_found.remove(&sp);
        }
    }
    if !not_found.is_empty() {
        return Err(E::InputConsistency(format!(
            "Species {1:?} cannot be found in input gene trees. {0} mispelled?",
            if not_found.len() == 1 { "Is it" } else { "Are they" },
            not_found
                .iter()
                .map(|&sp| ResolvedSymbol::new(sp, interner))
                .collect::<Vec<_>>(),
        )));
    }

    eprintln!("Echo input:");
    eprintln!("{:#?}", config.for_display(interner));
    eprintln!("  Found {} gene trees.", forest.len());

    //==============================================================================
    // Extract various information from the trees,
    // and handle data from them with some kind of 'dataframe' pattern:
    // an informal sequence of same-sized columns.
    let ids = forest.ids();
    let raw_trees = forest.trees();
    let raw_nodes = raw_trees.iter().map(GeneTree::len).collect::<Vec<_>>();
    let sequences_lengths = raw_trees
        .iter()
        .map(GeneTree::sequence_length)
        .collect::<Vec<_>>();

    //----------------------------------------------------------------------------------------------
    eprintln!("Prune trees so they only contain relevant species.");
    let (pruned_trees, pruned_nodes) = raw_trees.iter().fold(
        (Vec::new(), Vec::new()),
        |(mut trees, mut nodes), raw_tree| {
            let pruned = raw_tree.pruned(|_, leaf| relevant_species.contains(&leaf.payload));
            nodes.push(pruned.len());
            trees.push(pruned);
            (trees, nodes)
        },
    );

    //----------------------------------------------------------------------------------------------
    separator();
    eprintln!("Filter trees based on topology.\n");
    let mut n_excluded = 0;
    let mut n_unresolved = 0;

    // Leave None in place to keep same-sized columns.
    let mut passed_triplets = Vec::new();
    let mut passed_outgroup = Vec::new();
    let mut passed_top = Vec::new();
    let mut included = Vec::new();

    // If required, prepare for the second branches lengths based filter.
    let max_clock_ratio = config.filters.max_clock_ratio;
    let mut triplet_rest_sum_means = max_clock_ratio.is_some().then_some((0., 0.));
    let mut shapes = Vec::new();
    let mut shape_ratios = Vec::new();

    for i in 0..pruned_trees.len() {
        let pruned = &pruned_trees[i];
        let id = ids[i];
        let raw = &raw_trees[i];

        // Decide first filter based on topology.
        let topology = pruned.topological_analysis(&config.taxa);
        let pass = topology.is_included(config.filters.triplet_other_monophyly);
        n_excluded += usize::from(!pass);

        let (triplet, outgroup, treetop) = display_topological_tree_analysis(
            id,
            raw,
            pruned,
            &config.filters,
            &topology,
            pass,
            interner,
        );

        // Prepare possible second filter pass by calculating geometrical properties.
        if let Some((triplet_means_sum, rest_means_sum)) = &mut triplet_rest_sum_means {
            shapes.push(pass.then(|| {
                let (mean_triplet, mean_rest) = pruned.triplet_rest_mean_lengths(&topology);
                *triplet_means_sum += mean_triplet;
                *rest_means_sum += mean_rest;
                (mean_triplet, mean_rest)
            }));
        }

        // If the triplet filter passed, analyze its topology wrt species tree.
        let triplet = triplet.then(|| {
            let trip = extract_local_triplet(
                pruned,
                &topology,
                &config.taxa.triplet,
                config.unresolved_length,
            );
            if config.unresolved_length.is_some() && trip.topology.is_none() {
                n_unresolved += 1;
            }
            trip
        });

        passed_triplets.push(triplet);
        passed_outgroup.push(outgroup);
        passed_top.push(treetop);
        included.push(pass.then_some(pruned));
    }

    eprintln!(
        "  --> {n_excluded} tree{} excluded from analysis based on their topology.",
        if n_excluded > 1 { "s were" } else { " was" }
    );
    if let Some(min) = config.unresolved_length {
        eprintln!(
            "  --> {n_unresolved} included triplet{} unresolved (internal length <= {min}).",
            if n_unresolved > 1 { "s were" } else { " was" }
        );
    }

    //----------------------------------------------------------------------------------------------
    // Geometry filter.

    // Is is actually needed?
    let geometry_pass = max_clock_ratio.is_some();

    if geometry_pass {
        separator();
        eprintln!("Filter trees based on geometry.\n");
        let (sum_mean_triplets, sum_mean_rest) = triplet_rest_sum_means.unwrap();
        let max = max_clock_ratio.unwrap();

        let (triplet_longer, imbalance) = imbalance(sum_mean_triplets, sum_mean_rest);
        let (t, r) = ("'triplet' section", "'outgroup' and 'other' section");
        let (small, large) = if triplet_longer { (t, r) } else { (r, t) };
        eprintln!(
            "  Branch lengths in {large} are on average {imbalance} times longer \
               than in {small}."
        );
        let global_shape = sum_mean_triplets / sum_mean_rest;
        eprintln!("  Mean tree shape: {global_shape}");

        let mut n_excluded = 0;
        for i in 0..shapes.len() {
            if let Some(gtree) = included[i] {
                let id = ids[i];
                let (mean_triplet, mean_rest) = shapes[i].unwrap();
                let (pass, shape_ratio) =
                    gtree.is_shape_included(mean_triplet, mean_rest, global_shape, max);
                display_geometrical_tree_analysis(
                    id,
                    gtree,
                    (mean_triplet, mean_rest),
                    global_shape,
                    max,
                    pass,
                    interner,
                );
                if !pass {
                    included[i] = None;
                    n_excluded += 1;
                }
                shape_ratios.push(Some((pass, shape_ratio)));
            } else {
                shape_ratios.push(None);
            }
        }
        eprintln!(
            "  --> {n_excluded} tree{} excluded from analysis based on their geometry.",
            if n_excluded > 1 { "s were" } else { " was" }
        );
    }

    //----------------------------------------------------------------------------------------------
    separator();
    eprintln!("Calculate mean branches lengths per filtered tree.");
    let mean_lengths = included
        .iter()
        .map(|tree| tree.map(|t| t.mean_branch_length(t.root())))
        .collect::<Vec<_>>();

    eprintln!("Estimate relative mutation rates.");
    let global_mean = mean_lengths
        .iter()
        .filter_map(Option::clone)
        .summed_mean::<usize>();
    eprintln!("  mean branch length accross trees: l = {global_mean}");
    let mutation_rates = mean_lengths
        .iter()
        .map(|o| o.map(|local_mean| local_mean / global_mean))
        .collect::<Vec<_>>();

    //==============================================================================
    // Display as a table.
    separator();
    eprintln!("Pre-analysis results:\n");

    // Column names and alignment.
    let mut columns = vec![
        ("name", '<'),
        ("bases", '>'),
        ("raw", '>'),
        ("pruned", '>'),
        ("trip", '^'),
        ("out", '^'),
        ("top", '^'),
    ];
    if max_clock_ratio.is_some() {
        columns.push(("sratio", '>'));
    }
    columns.extend_from_slice(&[("l_i", '>'), ("α_i", '>')]);
    let (header, aligns): (Vec<_>, Vec<_>) = columns.into_iter().unzip();

    // First pass to prepare columns widths and content.
    let mut lines = Vec::new();
    let mut widths = header.iter().copied().map(str::len).collect::<Vec<_>>();
    for i in 0..ids.len() {
        let id = ids[i];
        let sl = &sequences_lengths[i];
        let n = &raw_nodes[i];
        let np = &pruned_nodes[i];
        let otrip = &passed_triplets[i];
        let out = passed_outgroup[i];
        let top = passed_top[i];
        let ml = mean_lengths[i];
        let mu = mutation_rates[i];

        let ok = "✔".green();
        let nok = "🗙".red();
        let filter = |b| (if b { &ok } else { &nok }).string();
        let mut line = vec![
            format!("{:?}", ResolvedSymbol::new(id, interner)),
            format!("{sl}"),
            format!("{n}"),
            format!("{np}"),
            if let Some(trip) = otrip {
                if let Some(top) = trip.topology {
                    format!("{top}")
                } else {
                    "<?>".yellow().string()
                }
            } else {
                nok.string()
            },
            filter(out),
            if let Some(f) = top { filter(f) } else { "-".into() },
        ];
        if max_clock_ratio.is_some() {
            line.push(if let &Some((pass, sr)) = &shape_ratios[i] {
                let sr = format!("{sr:.3}");
                if pass {
                    sr
                } else {
                    let sr = sr.italic().red();
                    format!("{nok} {sr}")
                }
            } else {
                "-".into()
            });
        }
        line.extend_from_slice(&[
            if let Some(ml) = ml { format!("{ml:.3}") } else { "-".into() },
            if let Some(mu) = mu { format!("{mu:.3}") } else { "-".into() },
        ]);
        for ((w, h), field) in widths.iter_mut().zip(&header).zip(&line) {
            *w = max(max(*w, h.len()), terminal_width(field));
        }
        lines.push(line);
    }

    // Second pass to sort by tree name.
    lines.sort_unstable_by(|a, b| natural_lexical_cmp(&a[0], &b[0]));

    // Third pass to actually display the table.
    // Header line.
    eprintln!(
        "|{}",
        header
            .iter()
            .zip(&widths)
            .fold(String::new(), |mut s, (&h, &w)| {
                write!(s, "{}|", justify(h, w + 2, " ", '^')).unwrap();
                s
            })
    );
    // Separator line.
    eprintln!(
        "|{}|",
        widths
            .iter()
            .map(|w| format!("{:-^a$}", "", a = w + 2))
            .collect::<Vec<String>>()
            .join("+")
    );
    // Data lines.
    for line in lines {
        eprintln!(
            "|{}",
            line.iter()
                .zip(&aligns)
                .zip(&widths)
                .fold(String::new(), |mut s, ((f, &a), &w)| {
                    write!(s, " {} |", justify(f, w, " ", a)).unwrap();
                    s
                })
        );
    }

    //----------------------------------------------------------------------------------------------
    eprintln!("\nSummary:");

    let s = |n| if n > 1 { "s" } else { "" };
    let n = passed_triplets
        .iter()
        .map(|g| u64::from(g.is_none()))
        .sum::<u64>();
    eprintln!(
        "  - {} triplet{} rejected (incomplete or non-monophyletic).",
        n,
        s(n)
    );
    if let Some(min) = config.unresolved_length {
        let n = passed_triplets
            .iter()
            .filter_map(|g| g.as_ref().map(|g| u64::from(g.topology.is_none())))
            .sum::<u64>();
        eprintln!(
            "  - {} triplet{} considered unresolved (internal branch length ⩽ {min}).",
            n,
            s(n)
        );
    }
    let n = passed_outgroup
        .iter()
        .map(|pass| u64::from(!pass))
        .sum::<u64>();
    eprintln!(
        "  - {} outroup{} rejected (empty or non-monophyletic).",
        n,
        s(n)
    );
    let n = passed_top
        .iter()
        .map(|&pass| if let Some(p) = pass { u64::from(!p) } else { 1 })
        .sum::<u64>();
    eprintln!(
        "  - {} tree top{} rejected (non-root LCA(triplet, outgroup){}).",
        n,
        s(n),
        if config.filters.triplet_other_monophyly {
            " or non-monophyletic (triplet, other)"
        } else {
            ""
        }
    );
    if let Some(max) = max_clock_ratio {
        let n = shape_ratios
            .iter()
            .filter_map(|o| o.map(|(pass, _)| u64::from(!pass)))
            .sum::<u64>();
        eprintln!(
            "  - {} tree shape{} rejected \
             (triplet/outgroup imbalance larger than {max} times the average).",
            n,
            s(n)
        );
    }
    let n = included.iter().map(|o| u64::from(o.is_some())).sum::<u64>();
    eprintln!("  => {} tree{} kept for analysis.", n, s(n));

    //==============================================================================
    separator();

    // Reduce number of lines in the "dataframe" to only focus on kept trees.
    let mut kept_ids = Vec::new();
    let mut triplets = Vec::new();
    for i in 0..ids.len() {
        if included[i].is_some() {
            let local = passed_triplets[i].take().unwrap();
            let &relative_mutation_rate = mutation_rates[i].as_ref().unwrap();

            // Upgrade local triplets to extended versions, aware of the whole dataset.
            let triplet = GeneTriplet {
                local,
                relative_mutation_rate,
            };

            kept_ids.push(ids[i]);
            triplets.push(triplet);
        }
    }
    drop(passed_triplets); // Has been consumed into.

    eprintln!("Starting point: {:#?}", config.init_parms);

    let parms = config.init_parms;
    eprintln!("\nInitial ln-likelihood:");
    let display_tree_lnl_detail = |parms, lnl| {
        for (&id, trip) in kept_ids.iter().zip(triplets.iter()) {
            eprintln!(
                "  {}\t{}",
                trip.likelihood(parms).ln(),
                interner.resolve(id).unwrap(),
            );
        }
        eprintln!("total: {lnl}");
    };
    let lnl = ln_likelihood(&triplets, &parms);
    display_tree_lnl_detail(&parms, lnl);

    eprintln!("\nOptimizing ln-likelihood:");
    let (opt, opt_lnl) = optimize_likelihood(&triplets, &parms);

    eprintln!("\nOptimized ln-likelihood:");
    display_tree_lnl_detail(&opt, opt_lnl);

    eprintln!("\nEnding point: {opt:#?}\n");
    Ok(())
}

// Convenience for coloring owned strings.
trait ToString {
    fn string(&self) -> String;
}
impl ToString for ColoredString {
    fn string(&self) -> String {
        format!("{self}")
    }
}
