// Verify raw config consistency as we construct actual config from it.
// Take this opportunity to intern all species names and replace them with their symbols.

use core::fmt;
use std::{
    cmp::Ordering,
    collections::HashSet,
    path::{Path, PathBuf},
};

use arrayvec::ArrayVec;
use regex::Regex;

use super::deserialize::{
    InitialParameters, NodeInput, RawConfig, RawFilters, RawGeneFlowTimes, RawTaxa, RawTriplet,
};
use crate::{
    gene_tree::{BranchLength, MeanNbBases},
    interner::{Interner, ResolvedSymbol, SpeciesSymbol},
    io::read_file,
    model::parameters::{GeneFlowTimes, MAX_N_GF_TIMES},
    Error as E, Error, Parameters,
};

// The final value handed out to user.
pub struct Config {
    // Path to the gene trees data file.
    pub trees: PathBuf,

    // Names of the species of interest.
    pub taxa: Taxa,

    // Starting point in the search space.
    pub init_parms: Parameters<f64>,

    // Parametrize the filters.
    pub filters: Filters,

    // If set, when the internal triplet branch length "d"
    // is (non-strictly) less than this value,
    // consider that its topology is *unresolved*.
    // In this situation, every scenario contributes to the likelihood,
    // because the topology "maybe matches",
    // and it is considered that "observed d = 0".
    // The value is given in *mutation* units, so branch length × sequence length.
    pub unresolved_length: Option<MeanNbBases>,
}

#[derive(Debug)]
pub struct Filters {
    // When raised, filter out trees if some species in 'other'
    // branche between LCA(outgroup) and the root,
    // because they may should have been considered 'outgroup' instead.
    pub triplet_other_monophyly: bool,

    // Let `mt` be the mean branch length for gene tree triplet species,
    // and `mr` ("rest") be the mean branch length for outgroup/other species.
    // The "shape" of the gene tree is the ratio `q = mt/mr`.
    // The "global shape" or "mean shape" of all gene trees is `Q = smt/smr`,
    // where `smt` is the sum of all `mt` and `smr` of all `mr`
    // for all trees *with accepted topology*.
    // If set, this parameter reject trees whose shape differ from the mean shape
    // by more than a factor, say `2`, or anything specified with this parameter.
    // The motivation is that the hypothesis that mutation rate is constant
    // accross the tree(s) is too audacious when, eg. `q / Q ⩾ max_clock_ratio`.
    // Values need to be greater than 1
    // because they quantify "absolute" ratio or "imbalance": `max(q/Q, Q/q)`.
    pub max_clock_ratio: Option<BranchLength>,
}

const DEFAULT_FILTERS: Filters = Filters {
    max_clock_ratio: None,
    triplet_other_monophyly: false,
};

// Contains only non-duplicated interned strings.
// Every section contains distinct species.
pub struct Taxa {
    pub triplet: SpeciesTriplet,
    pub(crate) outgroup: Vec<SpeciesSymbol>,
    pub(crate) other: Vec<SpeciesSymbol>,
}

// The triplet is importantly *oriented*: ((A, B), C).
// Accept several input forms provided they are unambiguous:
//    triplet = [["U", "V"], "W"] # A: U, B: V, C: W
//    triplet = ["U", ["V", "W"]] # A: V, B: W, C: U
// Or, shorter without special chars in species names:
//    triplet = "((U, V), W)"
//    triplet = "(U, (V, W))"
pub struct SpeciesTriplet {
    pub a: SpeciesSymbol,
    pub b: SpeciesSymbol,
    pub c: SpeciesSymbol,
}

macro_rules! err {
    ($($message:tt)*) => {{
        return Err(Error::InputConsistency(format!$($message)*));
    }};
}

impl Config {
    // Read raw config from file, then check it and convert into a valid config value.
    pub fn from_file(path: &Path, interner: &mut Interner) -> Result<Self, Error> {
        // Parse raw TOML file.
        let input = read_file(path)?;
        let raw = RawConfig::parse(&input)?;

        if let Some(ul) = raw.unresolved_length {
            if ul < 0. {
                err!(
                    ("The minimum internal branch length for tree considered resolved \
                      cannot be negative. Received 'unresolved_length = {ul}'.")
                )
            }
        }

        // Most checks implemented within `TryFrom` trait.
        Ok(Config {
            taxa: raw.taxa._try_into(interner)?,
            trees: raw.taxa.trees,
            init_parms: Parameters::try_from(&raw.initial_parameters, &raw.gf_times)?,
            unresolved_length: raw.unresolved_length,
            filters: if let Some(ref raw) = raw.filters {
                raw.try_into()?
            } else {
                DEFAULT_FILTERS
            },
        })
    }

    // The union of all sections constitutes the set of "relevant" species.
    pub fn relevant_species(&self) -> impl Iterator<Item = SpeciesSymbol> + '_ {
        let Taxa {
            triplet,
            outgroup,
            other,
        } = &self.taxa;
        triplet
            .iter()
            .chain(outgroup.iter().copied())
            .chain(other.iter().copied())
    }
}

//--------------------------------------------------------------------------------------------------
// Check filter parameters.
impl TryFrom<&RawFilters> for Filters {
    type Error = Error;

    fn try_from(raw: &RawFilters) -> Result<Self, Self::Error> {
        if let Some(mcr) = raw.max_clock_ratio {
            if mcr < 1. {
                err!(
                    ("The maximum branch length ratio between triplet and outgroup \
                      cannot be lower than 1. Received 'max_clock_ratio = {mcr}'.")
                )
            }
        }
        Ok(Filters {
            max_clock_ratio: raw.max_clock_ratio,
            triplet_other_monophyly: raw
                .triplet_other_monophyly
                .unwrap_or(DEFAULT_FILTERS.triplet_other_monophyly),
        })
    }
}

//--------------------------------------------------------------------------------------------------
// Check initial parameters.

impl Parameters<f64> {
    fn try_from(init: &InitialParameters, raw_gft: &RawGeneFlowTimes) -> Result<Self, Error> {
        use Ordering as O;
        let &InitialParameters {
            theta,
            tau_1,
            tau_2,
            p_ab,
            p_ac,
            p_bc,
            p_ancient_gf,
        } = init;

        macro_rules! check_positive {
            ($($name:ident),+) => {$({
                let value = init.$name;
                if value < 0. {
                    err!((
                        "Model parameter '{}' must be positive, not {value}.",
                        stringify!($name)
                    ))
                }
            })+};
        }
        macro_rules! check_prob {
            ($($name:ident),+) => {$({
                let value = init.$name;
                if !(0. ..=1.).contains(&value) {
                    err!((
                        "Model parameter '{}' must be a probability (between 0 and 1), not {value}.",
                        stringify!($name)
                    ))
                }
            })+};
        }

        check_positive!(theta, tau_1, tau_2);
        check_prob!(p_ab, p_ac, p_bc, p_ancient_gf);

        match tau_2.total_cmp(&tau_1) {
            O::Less => {
                err!(
                    ("The second coalescence time must be older than the first: \
                  maybe tau_1 ({tau_1}) and tau_2 ({tau_2}) were meant to be reversed?")
                )
            }
            O::Equal => {
                err!(
                    ("The two coalescence times cannot be identical: \
                  here tau_1 == tau_2 == {tau_1}.")
                )
            }
            O::Greater => {}
        };

        let s = p_ab + p_bc + p_ac;
        if 1. < s {
            err!(
                ("The sum of gene flow transfer probabilities must not be larger than 1. \
                  p_ab + p_bc + p_ac = {p_ab} + {p_bc} + {p_ac} = {s} > 1.")
            )
        }

        Ok(Parameters {
            theta,
            tau_1,
            tau_2,
            p_ab,
            p_ac,
            p_bc,
            p_ancient: p_ancient_gf,
            gf_times: raw_gft.try_into()?,
        })
    }
}

//--------------------------------------------------------------------------------------------------
// Check gene flow times.

impl TryFrom<&RawGeneFlowTimes> for GeneFlowTimes<f64> {
    type Error = Error;
    fn try_from(raw: &RawGeneFlowTimes) -> Result<Self, Self::Error> {
        if raw.is_empty() {
            err!(("No gene flow time specified."));
        }

        if raw.len() > MAX_N_GF_TIMES {
            err!((
                "Aphid does not support more than {MAX_N_GF_TIMES:?} gene flow times, \
                 but {} times were provided.",
                raw.len()
            ));
        }

        // Check values while sorting them.
        let mut gf_times = Self::new();
        for &time in raw {
            if time.is_infinite() || time.is_nan() || time < 0. {
                err!(("Invalid gene flow time: {time}."));
            }
            if time > 1.0 {
                err!(
                    ("Gene flow time cannot yet be specified earlier \
                      than latest species divergence, but {time} is greater than 1.")
                );
            }
            match gf_times
                .0
                .binary_search_by(|&t| t.total_cmp(&time).reverse())
            {
                Ok(_) => err!(("Gene flow time '{time}' specified twice.")),
                Err(i) => gf_times.0.insert(i, time),
            }
        }
        Ok(gf_times)
    }
}

//--------------------------------------------------------------------------------------------------
impl RawTaxa {
    fn _try_into(&self, interner: &mut Interner) -> Result<Taxa, Error> {
        // Check taxa consistency.
        let Self {
            triplet: raw_triplet,
            outgroup,
            other,
            ..
        } = self;
        let triplet: [&str; 3] = raw_triplet.try_into().map_err(E::InputConsistency)?;
        if outgroup.is_empty() {
            err!(("No outgroup species specified."));
        }

        // No duplicates should be found among the taxa explicitly given.
        let mut relevant_species = HashSet::new();
        let mut new_species = move |name, section| {
            if relevant_species.contains(&name) {
                err!(
                    ("Species {name:?} found in '{section}' section \
                      is given twice within the [taxa] table.")
                )
            }
            let symbol = interner.get_or_intern(name);
            relevant_species.insert(name);
            // Also record into the interner.
            Ok(symbol)
        };

        let [a, b, c] = triplet.map(|s| new_species(s, "triplet"));
        Ok(Taxa {
            triplet: SpeciesTriplet {
                a: a?,
                b: b?,
                c: c?,
            },
            outgroup: outgroup
                .iter()
                .map(|s| new_species(s, "outgroup"))
                .collect::<Result<Vec<_>, _>>()?,
            other: other
                .iter()
                .map(|s| new_species(s, "other"))
                .collect::<Result<Vec<_>, _>>()?,
        })
    }
}

//--------------------------------------------------------------------------------------------------
// Check triplet and extract strings in canonical order.

impl<'t> TryFrom<&'t RawTriplet> for [&'t str; 3] {
    type Error = String;

    // Work towards canonicalized ((a, b), c),
    // (x, y) refer to input nodes still undetermined.
    #[allow(clippy::many_single_char_names)]
    fn try_from(input: &'t RawTriplet) -> Result<Self, Self::Error> {
        use NodeInput as N;
        use RawTriplet as T;
        macro_rules! err {
            ($($message:tt)+) => {{
                return Err(format!$($message)+);
            }};
        }
        match input {
            T::Arrays(ars) => {
                let original = format!("{ars:?}");
                // Special-case one useful error.
                if ars.len() == 3 {
                    let mut species = ArrayVec::<&str, 3>::new();
                    for i in ars {
                        match i {
                            N::Leaf(sp) => species.push(sp),
                            N::Internal(_) => break,
                        }
                    }
                    if species.len() == 3 {
                        let [a, b, c] = species.into_inner().unwrap();
                        err!(
                            ("Ambiguous triplet:\n  {original}\nSpecify either\n  \
                              [[{a:?}, {b:?}], {c:?}]\nor\n   [{a:?}, [{b:?}, {c:?}]]")
                        );
                    }
                }
                // Fallback error for other invalid input types.
                macro_rules! invalid { () => {{
                    err!(
                        ("Invalid triplet specification. Expected 3 nested nodes under the form \
                         [[\"A\", \"B\"], \"C\"] or [\"C\", [\"A\", \"B\"]], got {original}.")
                    );
                }};}
                if ars.len() != 2 {
                    invalid!()
                }
                let mut it = ars.iter();
                let (x, y) = (it.next().unwrap(), it.next().unwrap());
                let ((N::Internal(ab), N::Leaf(c)) | (N::Leaf(c), N::Internal(ab))) = (x, y) else {
                    invalid!()
                };
                if ab.len() != 2 {
                    invalid!()
                }
                let mut it = ab.iter();
                let (a, b) = (it.next().unwrap(), it.next().unwrap());
                Ok([a, b, c])
            }
            T::Parenthesized(input) => {
                // A small tree with 3 expected nodes can be just regular-parsed.
                // Capture groups for trimmed leaf nodes.
                let specials = ",()"; // Forbidden in species names then.
                let mut it = "abc"
                    .chars()
                    .map(|x| format!(r"\s*(?<{x}>[^{specials}]+?)\s*"));
                let (a, b, c) = (it.next().unwrap(), it.next().unwrap(), it.next().unwrap());
                // Match internal node.
                let ab = format!(r"\s*\({a},{b}\)\s*");
                // There are two possible forms.
                let [alt1, alt2] = [format!("{ab},{c}"), format!("{c},{ab}")]
                    .map(|p| Regex::new(&format!(r"\s*\({p}\)\s*")).unwrap());
                // Match either for the whole triplet.
                let caps = if let Some(caps) = alt1.captures(input) {
                    caps
                } else if let Some(caps) = alt2.captures(input) {
                    caps
                } else {
                    err!(
                        ("Invalid triplet specification. Expected 3 nested nodes under the form \
                         \"((A, B), C)\" or \"(C, (A, B))\", got {input:?}.")
                    )
                };
                let extr = |x| caps.name(x).unwrap().as_str();
                Ok([extr("a"), extr("b"), extr("c")])
            }
        }
    }
}

impl SpeciesTriplet {
    pub fn as_array(&self) -> [SpeciesSymbol; 3] {
        [self.a, self.b, self.c]
    }

    pub fn iter(&self) -> impl Iterator<Item = SpeciesSymbol> {
        self.as_array().into_iter()
    }
}

//--------------------------------------------------------------------------------------------------
// Display.

// We cannot simply derive(Debug) for the config
// because a reference to the interner is necessary to resolve the species symbols into names.
// TODO: One unfortunate consequence is that the list of all fields needs to be mirrored here.
pub struct ConfigDisplay<'i> {
    config: &'i Config,
    interner: &'i Interner,
}

struct TaxaDisplay<'i> {
    taxa: &'i Taxa,
    interner: &'i Interner,
}

struct TripletDisplay<'i> {
    triplet: &'i SpeciesTriplet,
    interner: &'i Interner,
}

impl<'i> fmt::Debug for ConfigDisplay<'i> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.debug_struct("Config")
            .field("trees", &self.config.trees)
            .field("taxa", &self.config.taxa.for_display(self.interner))
            .field("init_parms", &self.config.init_parms)
            .field("filters", &self.config.filters)
            .field("unresolved_length", &self.config.unresolved_length)
            .finish()
    }
}

impl<'i> fmt::Debug for TaxaDisplay<'i> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let Taxa {
            triplet,
            outgroup,
            other,
        } = self.taxa;
        let resolve = |slice: &[_]| {
            slice
                .iter()
                .map(|&symbol| ResolvedSymbol::new(symbol, self.interner))
                .collect::<Vec<_>>()
        };
        f.debug_struct("Taxa")
            .field("triplet", &triplet.for_display(self.interner))
            .field("outgroup", &resolve(outgroup))
            .field("other", &resolve(other))
            .finish()
    }
}

// Display as [[A, B], C].
impl<'i> fmt::Debug for TripletDisplay<'i> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let [a, b, c] = self
            .triplet
            .as_array()
            .map(|s| ResolvedSymbol::new(s, self.interner));
        let ab = [a, b];
        f.debug_list().entry(&ab).entry(&c).finish()
    }
}

impl Config {
    pub fn for_display<'d>(&'d self, interner: &'d Interner) -> ConfigDisplay<'d> {
        ConfigDisplay {
            config: self,
            interner,
        }
    }
}

impl Taxa {
    fn for_display<'d>(&'d self, interner: &'d Interner) -> TaxaDisplay<'d> {
        TaxaDisplay {
            taxa: self,
            interner,
        }
    }
}

impl SpeciesTriplet {
    fn for_display<'d>(&'d self, interner: &'d Interner) -> TripletDisplay<'d> {
        TripletDisplay {
            triplet: self,
            interner,
        }
    }
}

//-------------------------------------------------------------------------------------------------
#[cfg(test)]
mod tests {
    use serde::Deserialize;

    use super::*;
    #[test]
    #[allow(clippy::too_many_lines)]
    fn triplet_parsing() {
        #[derive(Deserialize)]
        struct Single {
            raw: RawTriplet,
        }
        let success = |input, expected| {
            println!("Checking success for: {input}.");
            let raw: Single = toml::from_str(&format!("raw = {input}")).unwrap();
            let actual: [&str; 3] = (&raw.raw).try_into().unwrap();
            assert_eq!(expected, actual);
        };
        success(r#""((U, V), W)""#, ["U", "V", "W"]);
        success(r#""(U, (V, W))""#, ["V", "W", "U"]);
        success(
            r#"" ( ( with space , pretty much ) , everywhere ) ""#,
            ["with space", "pretty much", "everywhere"],
        );
        success(r#"[["U", "V"], "W"]"#, ["U", "V", "W"]);
        success(r#"["U", ["V", "W"]]"#, ["V", "W", "U"]);

        let failure = |input, expected: String| {
            println!("Checking failure for: {input}.");
            let raw = match toml::from_str::<Single>(&format!("raw = {input}")) {
                Ok(r) => r,
                Err(e) => {
                    let actual = format!("{e}");
                    assert_eq!(expected, actual, "Invalid error message.");
                    return;
                }
            };
            let res: [&str; 3] = match (&raw.raw).try_into() {
                Ok(r) => r,
                Err(e) => {
                    let actual = e.to_string();
                    assert_eq!(expected, actual, "Invalid error message.");
                    return;
                }
            };
            panic!("Unexpected success with\n  {input:?}\nObtained:\n  {res:?}");
        };
        failure(
            r#""(U, V, W)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "(U, V, W)"."#
            )
            .into(),
        );
        failure(
            r#""(U, V, W, X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "(U, V, W, X)"."#,
            )
            .into(),
        );
        failure(
            r#""((U, V, W), X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "((U, V, W), X)"."#,
            )
            .into(),
        );
        failure(
            r#""((U, V), W, X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "((U, V), W, X)"."#,
            )
            .into(),
        );
        failure(
            r#""((U, V), W, X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "((U, V), W, X)"."#,
            )
            .into(),
        );
        // Special-cased error message.
        failure(
            r#"["U", "V", "W"]"#,
            concat!(
                r#"Ambiguous triplet:\n  ["U", "V", "W"]\n"#,
                r#"Specify either\n  [["U", "V"], "W"]\n"#,
                r#"or\n   ["U", ["V", "W"]]"#,
            )
            .replace(r"\n", "\n"),
        );
        failure(
            r#"["U", "V", "W", "X"]"#,
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got ["U", "V", "W", "X"]."#
            )
            .into(),
        );
        failure(
            r#"[["U", "V"], "W", "X"]"#,
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got [["U", "V"], "W", "X"]."#
            )
            .into(),
        );
        failure(
            r#"[["U", "V"], ["W", "X"]]"#,
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got [["U", "V"], ["W", "X"]]."#
            )
            .into(),
        );
        failure(
            r"[]",
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got []."#
            )
            .into(),
        );
        // Not a super error.. but go with it for now.
        failure(
            r#"[["U", ["V", "W"]]]"#,
            concat!(
                r#"TOML parse error at line 1, column 7\n"#,
                r#"  |\n"#,
                r#"1 | raw = [["U", ["V", "W"]]]\n"#,
                r#"  |       ^^^^^^^^^^^^^^^^^^^\n"#,
                r#"data did not match any variant of untagged enum RawTriplet\n"#,
            )
            .replace(r"\n", "\n"),
        );
    }
}
