// Deserialize raw config from file.
// This type is a faithful representation
// what the user wrote in the config file,
// and it then needs to be checked
// into the eventual config type handed to user.

use std::{array, fmt, path::PathBuf};

use serde::Deserialize;
use serde_with::{serde_as, FromInto};

use crate::Error;

//--------------------------------------------------------------------------------------------------
// Global config structure and defaults.

pub(super) type RawGeneFlowTimes = Vec<f64>;

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(super) struct RawConfig {
    pub(super) filters: Option<RawFilters>,
    pub(super) taxa: RawTaxa,
    #[serde(rename = "init", default = "default_initial_parameters")]
    pub(super) initial_parameters: InitialParameters,
    #[serde(default = "default_gf_times")]
    pub(super) gf_times: RawGeneFlowTimes,
    pub(super) unresolved_length: Option<f64>,
}

impl RawConfig {
    // Parse config file successfully
    // or abort the program with a useful error.
    pub(super) fn parse(input: &str) -> Result<Self, Error> {
        Ok(toml::from_str(input)?)
    }
}

//--------------------------------------------------------------------------------------------------
// Filters configuration.

#[serde_as]
#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(super) struct RawFilters {
    pub(super) max_clock_ratio: Option<f64>,
    pub(super) triplet_other_monophyly: Option<bool>,
}

//--------------------------------------------------------------------------------------------------
// Species data.

#[serde_as]
#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(super) struct RawTaxa {
    // Where to find the gene trees.
    pub(super) trees: PathBuf,

    pub(super) triplet: RawTriplet,

    #[serde_as(as = "FromInto<ListOfStrings>")]
    pub(super) outgroup: Vec<String>,

    #[serde_as(as = "FromInto<ListOfStrings>")]
    pub(super) other: Vec<String>,
}

// Guard against invalid inputs.
#[derive(Deserialize)]
#[serde(untagged)]
pub(super) enum RawTriplet {
    Arrays(Vec<NodeInput>),
    Parenthesized(String),
}

#[derive(Deserialize)]
#[serde(untagged)]
pub(super) enum NodeInput {
    Leaf(String),
    Internal(Vec<String>),
}

impl fmt::Debug for RawTriplet {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        use RawTriplet as T;
        match self {
            T::Arrays(ars) => write!(f, "{ars:?}"),
            T::Parenthesized(s) => write!(f, "{s:?}"),
        }
    }
}

impl fmt::Debug for NodeInput {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        use NodeInput as N;
        match self {
            N::Leaf(s) => write!(f, "{s:?}"),
            N::Internal(vec) => write!(f, "{vec:?}"),
        }
    }
}

//--------------------------------------------------------------------------------------------------
// Statistical model parameters.

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(super) struct InitialParameters {
    pub(super) theta: f64,
    pub(super) tau_1: f64,
    pub(super) tau_2: f64,
    pub(super) p_ab: f64,
    pub(super) p_ac: f64,
    pub(super) p_bc: f64,
    pub(super) p_ancient_gf: f64,
}

fn default_initial_parameters() -> InitialParameters {
    InitialParameters {
        theta: 0.005,
        tau_1: 0.002,
        tau_2: 0.005,
        p_ab: 0.3,
        p_ac: 0.3,
        p_bc: 0.3,
        p_ancient_gf: 0.7,
    }
}

fn default_gf_times() -> Vec<f64> {
    vec![1.]
}

//--------------------------------------------------------------------------------------------------
// Parsing utils.

// Lists of strings can either be specified as regular arrays like in:
//    ["a", "b", "c", "d"]
// Or as whitespace-separated strings, like in:
//     "a b c d"
// The latter is shorter,
// but less flexible as it does not allow whitespace in strings.
#[derive(Deserialize)]
#[serde(untagged)]
enum ListOfStrings {
    Regular(Vec<String>),
    Compact(String),
}

impl From<ListOfStrings> for Vec<String> {
    fn from(lof: ListOfStrings) -> Self {
        use ListOfStrings as L;
        match lof {
            L::Regular(vec) => vec,
            L::Compact(input) => input.split_whitespace().map(ToOwned::to_owned).collect(),
        }
    }
}

impl<const N: usize> TryFrom<ListOfStrings> for [String; N] {
    type Error = String;

    fn try_from(lof: ListOfStrings) -> Result<Self, Self::Error> {
        use ListOfStrings as L;
        let vec = match lof {
            L::Regular(vec) => vec,
            L::Compact(input) => input.split_whitespace().map(ToOwned::to_owned).collect(),
        };
        if vec.len() == N {
            let mut it = vec.into_iter();
            Ok(array::from_fn(|_| it.next().unwrap()))
        } else {
            Err(format!(
                "Expected {N} items in list of strings, found {}.",
                vec.len()
            ))
        }
    }
}
