// Coordinate toplevel failure modes.

use std::{
    fmt::{self, Display},
    io::Error as IOError,
    string::FromUtf8Error,
};

use colored::Colorize;
use toml::de::Error as TomlError;

use crate::lexer::Error as LexerError;

// Toplevel errors.
#[cfg_attr(test, derive(Debug))]
pub enum Error {
    // Errors while reading the text in input files.
    InputIo(IOError),
    Utf8(FromUtf8Error),
    // Error when parsing the main config file.
    ConfigParsing(TomlError),
    // Error when checking input consistency.
    InputConsistency(String),
    // Error while reading gene trees, located within the file.
    GeneTree { line: usize, error: LexerError },
    // Fallback/ general purpose error.
    Any(String),
}

// Delegate display to contained errors.
impl Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = format!("{}:", self.prefix());
        write!(f, "{} {}", prefix.bold().red(), self.message())
    }
}

impl Error {
    // Pick an error prefix for each.
    fn prefix(&self) -> &str {
        use Error as E;
        match self {
            E::InputIo(_) => "Input error",
            E::Utf8(_) => "Encoding error",
            E::ConfigParsing(_) => "Config parsing error",
            E::InputConsistency(_) => "Inconsistent input",
            E::GeneTree { .. } => "Gene tree error",
            E::Any { .. } => "Error",
        }
    }

    // Pick an error code for each.
    pub fn code(&self) -> i32 {
        use Error as E;
        match self {
            E::InputIo(_) => 1,
            E::Utf8(_) => 2,
            E::ConfigParsing(_) => 3,
            E::InputConsistency(_) => 4,
            E::GeneTree { .. } => 5,
            E::Any { .. } => 6,
        }
    }

    // Extract owned message string from whatever underlying error.
    fn message(&self) -> String {
        use Error as E;
        #[allow(clippy::match_same_arms)]
        match self {
            E::InputIo(e) => format!("{e}"),
            E::Utf8(e) => format!("{e}"),
            E::InputConsistency(e) => e.to_string(),
            E::Any(e) => e.to_string(),
            E::GeneTree { line, error: e } => format!("Line {line}: {e}"),
            E::ConfigParsing(e) => {
                // Best-attempt to improve over https://github.com/serde-rs/serde/issues/773.
                let mut message = format!("{e}");
                let pattern = "data did not match any variant of untagged enum ";
                if message.contains(pattern) {
                    let name = message.trim().rsplit(' ').next().unwrap();
                    let accepted_types = match name {
                        // TODO: generate these with procedural macros?
                        "ListOfStrings" => vec!["Array", "String"],
                        _ => panic!("Error in source code: enum {name} has not been handled."),
                    };
                    message = message.replace(
                        &format!("{pattern}{name}"),
                        &format!(
                            "invalid type: expected one of [{}]",
                            accepted_types.join(", ")
                        ),
                    );
                };
                message
            }
        }
    }
}

// Trivial conversions.
macro_rules! from {
    ($($From:ident -> $To:ident)*) => {$(
        impl From<$From> for Error {
            fn from(e: $From) -> Self {
                Error::$To(e)
            }
        }
    )*};
}

from! {
    IOError -> InputIo
    FromUtf8Error -> Utf8
    TomlError -> ConfigParsing
}
