// The GLS/ILF model, associated parameters and likelihood calculations.

use std::io::{self, Write};

use crate::GeneTriplet;

mod likelihood;
pub(crate) mod parameters;
mod scenarios;
mod scores;

pub use parameters::Parameters;
use tch::{
    nn::{self, Module, OptimizerConfig},
    Device, Kind, Tensor,
};

use self::likelihood::likelihood_tensors;

// Log-probability that all triplets in the data occur.
pub fn ln_likelihood(
    // Receive in this format for better locality.
    triplets: &[GeneTriplet],
    p: &Parameters<f64>,
) -> f64 {
    let mut ln_prob = 0.;
    for gtrip in triplets {
        ln_prob += gtrip.likelihood(p).ln();
    }
    ln_prob
}

// Different implementation with torch tensors.
// Return both optimized input and output.
#[allow(clippy::similar_names)] // p_ab/p_ac is okay.
pub fn optimize_likelihood(
    // Receive in this format for better locality.
    triplets: &[GeneTriplet],
    start: &Parameters<f64>,
) -> (Parameters<f64>, f64) {
    // Choose in-memory data location.
    let (kind, device) = (Kind::Double, Device::cuda_if_available());
    let vars = nn::VarStore::new(device);

    // Construct the likelihood function from fixed data and initial trainable parameters.
    let (scores, ln_likelihood) = likelihood_tensors(&vars.root(), triplets, start, (kind, device));
    // Dummy input for this function: the data does not change during optimisation.
    let input = Tensor::zeros([], (kind, device));

    // Pick an optimizer and associated learning rate.
    let mut opt = nn::Sgd::default().build(&vars, 1e-3).unwrap();
    let n_steps = 1_000;

    // Optimize.
    let display_step = |i, lnl: &Tensor| {
        eprint!("step {i:03}: {}\r", lnl.double_value(&[]));
        io::stdout().flush().unwrap();
    };
    for i in 0..n_steps {
        opt.zero_grad();
        let lnl = ln_likelihood.forward(&input);
        if i % 97 == 0 || i == n_steps {
            display_step(i, &lnl);
        }
        opt.backward_step(&-lnl);
    }

    // Final value.
    let lnl = ln_likelihood.forward(&input);
    display_step(n_steps, &lnl);
    eprintln!();

    let parms: Parameters<f64> = (&scores).into();
    (parms, lnl.double_value(&[]))
}
