// Calculate chances that the given tree happens given parameters values.

use std::iter;

use itertools::izip;
use paste::paste;
use statrs::distribution::{Discrete, Poisson};
use tch::{nn, Device, IndexOp, Kind, Tensor};

use super::{
    parameters::GeneFlowTimes,
    scenarios::{GeneFlowScenario, GeneFlowTopology, Scenario, N_ILS_SCENARIOS},
    scores::Scores,
    Parameters,
};
use crate::gene_tree::{nb_mutations, triplet::GeneTriplet};

//--------------------------------------------------------------------------------------------------
macro_rules! probabilities_ils_nongf {
    ($parameters:expr, $Ty:ty) => {{
        let Parameters {
            theta,
            tau_1,
            tau_2,
            p_ab,
            p_ac,
            p_bc,
            ..
        } = $parameters;
        let p_ils = ((2.0 * (tau_1 - tau_2) / theta) as $Ty).exp();
        let p_ngf = 1. - p_ab - p_ac - p_bc;
        [p_ils, p_ngf]
    }};
}
impl Parameters<f64> {
    fn probabilities_ils_nongf(&self) -> [f64; 2] {
        probabilities_ils_nongf!(self, f64)
    }
}
impl Parameters<Tensor> {
    fn probabilities_ils_nongf(&self) -> [Tensor; 2] {
        probabilities_ils_nongf!(self, Tensor)
    }
}

macro_rules! scenario_probability {
    ($scenario:expr, $parameters:expr
     $(, $Ty:ident)? // Weak hack to ease factorization in this very particular (f64, Tensor) abstraction.
     ) => {
        paste! {{
            let p = $parameters;
            let Parameters {
                theta,
                tau_1,
                tau_2,
                p_ab,
                p_ac,
                p_bc,
                ..
            } = p;
            {
                use Scenario as S;
                let p_ils = ((2.0 * (tau_1 - tau_2) / theta) $(as $Ty)?).exp();
                let p_ngf = 1. - p_ab - p_ac - p_bc;
                match $scenario {
                    S::NoEvent => (1. - p_ils) * p_ngf,
                    #[allow(clippy::cast_precision_loss)]
                    S::IncompleteLineageSorting(_) => (1. / N_ILS_SCENARIOS as f64) * p_ils * p_ngf,
                    S::GeneFlow(gf) => gf.[<probability $(_ $Ty:lower s)?>](p),
                }
            }
        }}
    };
}

// Prior probability that the scenario occurs,
// with respect to other possible scenarios.
impl Scenario {
    pub(crate) fn probability(&self, p: &Parameters<f64>) -> f64 {
        scenario_probability!(self, p)
    }
    pub(crate) fn probability_tensors(&self, p: &Parameters<Tensor>) -> Tensor {
        scenario_probability!(self, p, Tensor)
    }
}

macro_rules! geneflow_probability {
    ($geneflow:expr, $parameters:expr) => {{
        let Self { topology, time } = $geneflow;
        let Parameters {
            p_ab,
            p_ac,
            p_bc,
            p_ancient,
            ..
        } = $parameters;
        let ngf = $parameters.n_gf_times();
        {
            use GeneFlowTopology as GT;
            let prob = match topology {
                GT::ABC => 1. * p_ab, // Trivial multiplication for better typecheck.
                GT::ACB | GT::CAB => 0.5 * p_ac,
                GT::BCA | GT::CBA => 0.5 * p_bc,
            };
            #[allow(clippy::cast_precision_loss)]
            if (*time) == 0 {
                p_ancient * prob
            } else {
                (1. - p_ancient) * prob / (ngf - 1) as f64
            }
        }
    }};
}
impl GeneFlowScenario {
    pub(crate) fn probability(&self, p: &Parameters<f64>) -> f64 {
        geneflow_probability!(self, p)
    }
    pub(crate) fn probability_tensors(&self, p: &Parameters<Tensor>) -> Tensor {
        geneflow_probability!(self, p)
    }
}

//--------------------------------------------------------------------------------------------------
// Expected branches lengths for the triplet, according to the given scenario:
// Return only pair of short/long branches lengths (S/L).
//
//          ┌──┴──┐
//         d│L-S  │
//        ┌─┴─┐  c│L
//       a│S b│S  │
//
// Abstract over operand types,
// not with trait bounds because they would be too intricate.
macro_rules! branches_lengths {
    ($scenario:expr, $parameters:expr) => {{
        use Scenario as S;
        let Parameters {
            theta,
            tau_1,
            tau_2,
            gf_times,
            ..
        } = $parameters;
        let gf_times = |i| &gf_times.0[i];
        match $scenario {
            S::NoEvent => [tau_1 + 0.5 * theta, tau_2 + 0.5 * theta],
            S::IncompleteLineageSorting(_) => {
                [tau_2 + (1. / 6.) * theta, tau_2 + (2. / 3.) * theta]
            }
            &S::GeneFlow(GeneFlowScenario {
                time: i,
                ref topology,
            }) => {
                use GeneFlowTopology as G;
                let tau_g = tau_1 * gf_times(i);
                [
                    tau_g,
                    match topology {
                        G::ABC | G::CAB | G::CBA => tau_2,
                        G::ACB | G::BCA => tau_1,
                    } + 0.5 * theta,
                ]
            }
        }
    }};
}
impl Scenario {
    pub(crate) fn branches_lengths(&self, p: &Parameters<f64>) -> [f64; 2] {
        branches_lengths!(self, p)
    }
    pub(crate) fn branches_lengths_tensors(&self, p: &Parameters<Tensor>) -> [Tensor; 2] {
        // branches_lengths!(self, p)
        {
            use Scenario as S;
            let Parameters {
                theta,
                tau_1,
                tau_2,
                gf_times,
                ..
            } = p;
            let gf_times = |i| &gf_times.0[i];
            match self {
                S::NoEvent => [tau_1 + 0.5 * theta, tau_2 + 0.5 * theta],
                S::IncompleteLineageSorting(_) => {
                    [tau_2 + (1. / 6.) * theta, tau_2 + (2. / 3.) * theta]
                }
                &S::GeneFlow(GeneFlowScenario {
                    time: i,
                    ref topology,
                }) => {
                    use GeneFlowTopology as G;
                    let tau_g = tau_1 * gf_times(i);
                    [
                        tau_g,
                        match topology {
                            G::ABC | G::CAB | G::CBA => tau_2,
                            G::ACB | G::BCA => tau_1,
                        } + 0.5 * theta,
                    ]
                }
            }
        }
    }
}

// Probability that the given gene triplet occurs
// given the suggested [S, L] branches lengths.
impl GeneTriplet {
    #[allow(clippy::many_single_char_names)]
    fn branches_lengths_likelihood(&self, [s, l]: [f64; 2]) -> f64 {
        // (locally match the paper notations)
        let alpha_i = self.relative_mutation_rate;
        let loc = &self.local;

        // Minimize .exp() evaluations within poisson density by incursing into log space.
        let ln_pmf = |a, e| {
            // The expected branch length is not yet a number of mutations.
            let actual = u64::from(a);
            let expected = alpha_i * nb_mutations(e, loc.sequence_length);
            ln_poisson_density(actual, expected)
        };

        // Match every actual branch length against its own expected length.
        let [a, b, d /*~ab*/, c] = loc.branches_lengths;
        let pa = ln_pmf(a, s); // One short branch.
        let pb = ln_pmf(b, s); // The other short branch.
        let pc = ln_pmf(c, l); // The long, external branch.
        let pd = ln_pmf(d, l - s); // The internal branch.

        (pa + pb + pc + pd).exp()
    }
}

fn ln_poisson_density(n: u64, lambda: f64) -> f64 {
    Poisson::new(lambda).unwrap().ln_pmf(n)
}

//--------------------------------------------------------------------------------------------------
// Probability that the given gene triplet occurs,
// integrating over all possible scenarios.
impl GeneTriplet {
    pub fn likelihood(&self, p: &Parameters<f64>) -> f64 {
        // Only sum over concordant scenarios.
        let mut prob = 0.;
        for scenario in Scenario::iter(p.gf_times.0.len()) {
            if let Some(top) = self.local.topology {
                if top != scenario.topology() {
                    continue;
                }
            }
            let prob_scenario = scenario.probability(p);
            let expected_lengths = scenario.branches_lengths(p);
            let prob_branches = self.branches_lengths_likelihood(expected_lengths);
            prob += prob_scenario * prob_branches;
        }
        prob
    }
}

//==================================================================================================
// Alternate implementation as a torch module, for autograd + optimisation.

pub(crate) fn likelihood_tensors(
    p: &nn::Path,
    triplets: &[GeneTriplet],
    start: &Parameters<f64>,
    (kind, device): (Kind, Device),
) -> (Scores<Tensor>, impl nn::Module) {
    let n_trees = triplets.len();
    let ngf = start.n_gf_times();
    let scenarios = Scenario::iter(ngf).collect::<Vec<_>>();
    let n_scenarios = scenarios.len();

    let n_trees_i: i64 = n_trees
        .try_into()
        .unwrap_or_else(|_| panic!("Too many trees ({n_trees}) to be indexed with i64 by torch."));

    // Model log-likelihood as F(X, P) with X the 'data', constant throughout the optimisation,
    // and 'P' the 'parameters' to optimize, and that we wish to derive F with respect to.

    // Prepare all parameters 'P' from the desired initial value.
    let scores = leaves_tensors(p, start, device);

    // Extract all data 'X' from the triplets.
    let DataTensors {
        alphaseq_per_scenario,
        n_per_scenario,
        minus_ln_factorial_n_per_scenario,
        columns_indices,
    } = data_tensors(triplets, n_scenarios, &scenarios, device);

    // Calculate likelihood.
    (
        scores.alias(),
        nn::func(move |_| {
            // Calculate parameters for the current scores values.
            let parms: Parameters<Tensor> = (&scores).into();

            let mut columns = Vec::with_capacity(n_scenarios);

            for (scenario, alphaseq, n, minus_ln_factorial_n) in izip!(
                &scenarios,
                &alphaseq_per_scenario,
                &n_per_scenario,
                &minus_ln_factorial_n_per_scenario,
            ) {
                // Scenario prior probability.
                let prior = scenario.probability_tensors(&parms);

                // Expected branch lengths.
                let [ref s, ref l] = scenario.branches_lengths_tensors(&parms);
                let expected_abcd = Tensor::concatenate(&[s, s, &(l - s), l], 0);

                // Log-densities of poisson distribution.
                let lambda = Tensor::outer(alphaseq, &expected_abcd);
                let ln_poisson = minus_ln_factorial_n + n * lambda.log() - lambda;

                // Sum over the branches and exponentiate
                // to get the product of poisson distributions.
                let poisson_product = ln_poisson.sum_dim_intlist(1, false, kind).exp();

                // Down to the final contribution of every tree
                // to the likelihood of this scenario.
                let contributions = prior * poisson_product;

                columns.push(contributions);
            }

            // Sum over all scenarios to get the total probability of every tree.
            // This is a sum over the rows of the conceptual sparse matrix.
            let columns = Tensor::cat(&columns, 0);
            let trees_prob =
                Tensor::zeros(n_trees_i, (kind, device)).scatter_add(0, &columns_indices, &columns);

            // And finally integrate this all into the total likelihood.
            trees_prob.log().sum(None)
        }),
    )
}

// Construct all (potentially) gradient-tracked parameters,
// sources of the calculation graph towards the eventual likelihood.
fn leaves_tensors(p: &nn::Path<'_>, start: &Parameters<f64>, dev: Device) -> Scores<Tensor> {
    let single = |v, track| {
        Tensor::from_slice(&[v])
            .set_requires_grad(track)
            .to_device(dev)
    };
    let var = |v, name: &str, track| {
        if track {
            p.var_copy(name, &single(v, true))
        } else {
            single(v, false)
        }
    };

    // Calculate free-ranging scores from the desired starting point.
    let s: Scores<_> = start.into();
    let scores = Scores::<Tensor> {
        theta: var(s.theta, "theta", true),
        tau_1: var(s.tau_1, "tau_1", true),
        delta_tau: var(s.delta_tau, "delta_tau", true),
        gf: var(s.gf, "gf", true),
        ac: var(s.ac, "ac", true),
        bc: var(s.bc, "bc", true),
        ancient: var(s.ancient, "ancient", true),
        gf_times: GeneFlowTimes(
            s.gf_times
                .0
                .iter()
                .enumerate()
                .map(|(i, &v)| var(v, &format!("gf_time_{i:02}"), false))
                .collect(),
        ),
    };
    scores
}

// Construct all non(-potentially)-gradient-tracked tensors,
// referring to the immutable input data,
// and used to calculate the eventual likelihood.
fn data_tensors(
    triplets: &[GeneTriplet],
    n_scenarios: usize,
    scenarios: &[Scenario],
    dev: Device,
) -> DataTensors {
    let mut n = Vec::new();
    //                                     /!\
    // Generic name for branch lengths a/b/*d*/c.
    let mut alpha = Vec::new();
    let mut seqlen = Vec::new();
    for trip in triplets {
        alpha.push(trip.relative_mutation_rate);
        n.push(trip.local.branches_lengths.map(f64::from));
        seqlen.push(f64::from(trip.local.sequence_length));
    }
    let alpha = Tensor::from_slice(&alpha).to_device(dev);
    let n = Tensor::from_slice2(&n).to_device(dev);

    // Costly pre-calculation.
    let minus_ln_factorial_n = -(&n + 1).special_gammaln();
    let seqlen = Tensor::from_slice(&seqlen).to_device(dev);
    // The value actually used to calculate poisson lambda parameter.
    let alphaseq = alpha * seqlen;

    // The likelihood formula differs for every scenario,
    // later combined into a sum.
    // For every scenario, pick indices of the trees matching it.
    let mut match_trees: Vec<_> = iter::repeat_with(Vec::new).take(n_scenarios).collect();
    for (i_tree, trip) in triplets.iter().enumerate() {
        for (i_scenario, scenario) in scenarios.iter().enumerate() {
            if let Some(top) = trip.local.topology {
                if top != scenario.topology() {
                    continue;
                }
            }
            match_trees[i_scenario].push(i64::try_from(i_tree).unwrap());
        }
    }
    let scenario_trees_index = match_trees
        .into_iter()
        .map(|trees_indices| Tensor::from_slice(&trees_indices).to_device(dev))
        .collect::<Vec<_>>();

    // For every scenario, the immutable 'X' slice of relevant data
    // can be extracted for the trees concerned.
    let alphaseq_per_scenario = scenario_trees_index
        .iter()
        .map(|i_trees| alphaseq.i(i_trees))
        .collect::<Vec<_>>();
    let n_per_scenario = scenario_trees_index
        .iter()
        .map(|i_trees| n.i((i_trees, ..)))
        .collect::<Vec<_>>();
    let minus_ln_factorial_n_per_scenario = scenario_trees_index
        .iter()
        .map(|i_trees| minus_ln_factorial_n.i((i_trees, ..)))
        .collect::<Vec<_>>();

    // For every scenario, collect contribution of every tree to its likelihood,
    // in a sparse fashion corresponding to the columns of a conceptual scenario × gene trees matrix.
    // This matrix remains conceptual and is never actually reified into a tensor.
    let columns_indices = Tensor::cat(&scenario_trees_index, 0);
    DataTensors {
        alphaseq_per_scenario,
        n_per_scenario,
        minus_ln_factorial_n_per_scenario,
        columns_indices,
    }
}

// Return value of the above.
struct DataTensors {
    alphaseq_per_scenario: Vec<Tensor>,
    n_per_scenario: Vec<Tensor>,
    minus_ln_factorial_n_per_scenario: Vec<Tensor>,
    columns_indices: Tensor,
}
