// Most species in a gene tree are only useful to calculate useful statistics,
// before we actually focus the analysis on the triplet of species of interest.

use std::fmt::{self, Display};

use crate::{
    config::SpeciesTriplet,
    gene_tree::nb_mutations,
    topological_analysis::{SectionAnalysis, SectionLca},
    tree::Node,
    GeneTree, TopologicalAnalysis,
};

use super::{MeanNbBases, NbBases};

// Matches the gene triplet topology onto the expected species topology.
#[derive(Debug, Clone, Copy, PartialEq)]
#[allow(clippy::upper_case_acronyms)]
pub enum Topology {
    ABC, // ((A,B),C) -> Concordant triplets.
    ACB, // ((A,C),B) -> The outer node is B instead of C.
    BCA, // ((B,C),A) -> The outer node is A instead of C.
}

// Extract the minimal information needed from a full gene tree
// for likelihood analysis.
#[allow(clippy::module_name_repetitions)]
#[cfg_attr(test, derive(PartialEq, Debug))]
pub struct LocalGeneTriplet {
    //                                                           /!\
    // Rounded branches lengths, in canonicalized order: [{a, b}, d, {c}],
    // regardless of the contingent input gene tree ordering.
    //
    //   ┌──┴──┐  OR  ┌──┴──┐
    //   │d    │      │     │d
    // ┌─┴─┐   │c     │c  ┌─┴─┐
    // │a  │b  │      │   │a  │b
    //
    // Lengths stored here are not 'per-site' guesses,
    // they are a rounded integer estimate of the number of mutations.

    // If considered too small, the length 'd' is forced to zero,
    // and then the topology is considered 'unresolved':
    //
    //   ┌──┼──┐ d = 0
    //   │a │b │c
    //
    // In this situation, all scenarios are considered possible.

    // Notice that d comes *before* c, and the a/b order is arbitrary.
    //                                              /!\
    pub(crate) branches_lengths: [NbBases; 4], // [a, b, d, c]
    pub sequence_length: NbBases,
    // Concordance with respect to the species triplet.
    pub topology: Option<Topology>,
}

// Extend the above type with data
// which cannot be calculated from a tree alone,
// but from a reading through the whole forest.
#[allow(clippy::module_name_repetitions)]
pub struct GeneTriplet {
    pub local: LocalGeneTriplet,
    pub relative_mutation_rate: f64, // Known in the paper as "alpha_i".
}

// Assuming the gene tree has been correctly checked,
// match it to the associated topology.
// Refer to the species topology as ((a, b), c)
// and to the actual gene tree topology as ((u, v), w).
#[allow(clippy::many_single_char_names)]
pub fn extract_local(
    gtree: &GeneTree,
    analysis: &TopologicalAnalysis,
    species_triplet: &SpeciesTriplet,
    unresolved_length: Option<MeanNbBases>,
) -> LocalGeneTriplet {
    use Node as N;
    let err = || panic!("Did invalid trees pass the filter?");

    // Extract triplet ancestor.
    let SectionAnalysis::AllFound(SectionLca { id: lca_id, .. }) = analysis.triplet else {
        err()
    };
    let nodes = &gtree.tree.nodes;
    let N::Internal(lca) = &nodes[lca_id] else {
        err();
    };

    // Use it to extract triplet nodes, regardless of contingent tree ordering.
    let ([N::Internal(uv), N::Terminal(w)] | [N::Terminal(w), N::Internal(uv)]) =
        lca.children.map(|i| &nodes[i])
    else {
        err()
    };
    let [u, v] = uv.children.map(|i| {
        let N::Terminal(n) = &nodes[i] else { err() };
        n
    });

    // Possibly, the internal branch length is too small to be considered positive,
    // and the topology to be considered resolved.
    let resolved = if let Some(min) = unresolved_length {
        min < nb_mutations(uv.branch, gtree.sequence_length)
    } else {
        true
    };

    LocalGeneTriplet {
        branches_lengths: {
            // Extract branches lengths in canonicalized nodes order.
            let [u, v, w] = [u, v, w].map(|n| n.branch);
            let uv = if resolved { uv.branch } else { 0. };
            // Convert per-site estimates to total, sequence-wide estimates,
            // and round to an integer to match the Poisson distribution model.
            #[allow(clippy::cast_sign_loss, clippy::cast_possible_truncation)]
            [u, v, uv, w].map(|l| nb_mutations(l, gtree.sequence_length).round() as NbBases)
        },
        sequence_length: gtree.sequence_length,
        topology: resolved.then(|| {
            use Topology as T;
            // Compare to the species topology to figure the order.
            let [a, b, c] = species_triplet.as_array();
            // Only the outermost node is useful in this respect.
            let w = w.payload;
            if w == c {
                T::ABC
            } else if w == b {
                T::ACB
            } else if w == a {
                T::BCA
            } else {
                err()
            }
        }),
    }
}

impl Topology {
    pub(crate) fn iter() -> impl Iterator<Item = Self> {
        use Topology as T;
        [T::ABC, T::ACB, T::BCA].into_iter()
    }
}

impl Display for Topology {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        use Topology as T;
        f.write_str(match self {
            T::ABC => "ABC",
            T::ACB => "ACB",
            T::BCA => "BCA",
        })
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{config::Taxa, interner::Interner};

    #[test]
    #[allow(clippy::too_many_lines)] // It does take vertical space to test many trees.
    fn extract_triplet_from_gene_tree() {
        use Topology as T;
        let mut interner = Interner::new();
        let mut g = |c| interner.get_or_intern(c);

        let taxa = Taxa {
            triplet: SpeciesTriplet {
                a: g("T"),
                b: g("U"),
                c: g("V"),
            },
            outgroup: vec![g("O"), g("P"), g("Q")],
            other: vec![g("X"), g("Y"), g("Z")],
        };

        let unresolved_length = Some(0.5);

        let sequence_length = 10;
        let mut check = |input: &str, expected| {
            println!("Checking triplet extraction for tree:\n  {input}");
            let input = format!("{input} {sequence_length}");
            let gtree = GeneTree::from_input(&input, &mut interner).unwrap();
            let analysis = gtree.topological_analysis(&taxa);
            let actual = extract_local(&gtree, &analysis, &taxa.triplet, unresolved_length);
            assert_eq!(expected, actual);
        };

        // Tree with only a triplet.
        //
        //          │
        //       ┌──0──┐
        //       │     │
        //     ┌─1─┐   │
        //     2   3   4
        //     T   U   V
        check(
            "((T:2,U:3):1,V:4):0;",
            LocalGeneTriplet {
                branches_lengths: [20, 30, 10, 40],
                sequence_length,
                topology: Some(T::ABC),
            },
        );

        // Considered unresolved.
        for uv in [0., 0.03, 0.05] {
            check(
                &format!("((T:2,U:3):{uv},V:4):0;"),
                LocalGeneTriplet {
                    branches_lengths: [20, 30, 0, 40],
                    sequence_length,
                    topology: None,
                },
            );
        }
        // Considered resolved.
        let uv = 0.06;
        check(
            &format!("((T:2,U:3):{uv},V:4):0;"),
            LocalGeneTriplet {
                branches_lengths: [20, 30, 1, 40],
                sequence_length,
                topology: Some(T::ABC),
            },
        );

        //------------------------------------------------------------------------------------------
        // Synonymous concordant triplets.
        //
        //          │
        //       ┌──0──┐
        //       │     │
        //       │   ┌─2─┐
        //       1   3   4
        //       V   T   U
        check(
            "(V:1,(T:3,U:4):2):0;",
            LocalGeneTriplet {
                branches_lengths: [30, 40, 20, 10],
                sequence_length,
                topology: Some(T::ABC),
            },
        );

        // Considered unresolved.
        for uv in [0., 0.03, 0.05] {
            check(
                &format!("(V:1,(T:3,U:4):{uv}):0;"),
                LocalGeneTriplet {
                    branches_lengths: [30, 40, 0, 10],
                    sequence_length,
                    topology: None,
                },
            );
        }
        // Considered resolved.
        let uv = 0.06;
        check(
            &format!("(V:1,(T:3,U:4):{uv}):0;"),
            LocalGeneTriplet {
                branches_lengths: [30, 40, 1, 10],
                sequence_length,
                topology: Some(T::ABC),
            },
        );

        //
        //          │
        //       ┌──0──┐
        //       │     │
        //       │   ┌─2─┐
        //       1   3   4
        //       V   U   T
        check(
            "(V:1,(U:3,T:4):2):0;",
            LocalGeneTriplet {
                branches_lengths: [30, 40, 20, 10],
                sequence_length,
                topology: Some(T::ABC),
            },
        );

        //          │
        //       ┌──0──┐
        //       │     │
        //     ┌─1─┐   │
        //     2   3   4
        //     U   T   V
        check(
            "((U:2,T:3):1,V:4):0;",
            LocalGeneTriplet {
                branches_lengths: [20, 30, 10, 40],
                sequence_length,
                topology: Some(T::ABC),
            },
        );

        //------------------------------------------------------------------------------------------
        // Discordant triplets.
        //
        //          │
        //       ┌──0──┐
        //       │     │
        //       │   ┌─2─┐
        //       1   3   4
        //       T   V   U
        check(
            "(T:1,(V:3,U:4):2):0;",
            LocalGeneTriplet {
                branches_lengths: [30, 40, 20, 10],
                sequence_length,
                topology: Some(T::BCA),
            },
        );

        //
        //          │
        //       ┌──0──┐
        //       │     │
        //       │   ┌─2─┐
        //       1   3   4
        //       U   V   T
        check(
            "(U:1,(V:3,T:4):2):0;",
            LocalGeneTriplet {
                branches_lengths: [30, 40, 20, 10],
                sequence_length,
                topology: Some(T::ACB),
            },
        );

        //
        //          │
        //       ┌──0──┐
        //       │     │
        //     ┌─1─┐   │
        //     2   3   4
        //     U   V   T
        check(
            "((U:2,V:3):1,T:4):0;",
            LocalGeneTriplet {
                branches_lengths: [20, 30, 10, 40],
                sequence_length,
                topology: Some(T::BCA),
            },
        );

        //
        //          │
        //       ┌──0──┐
        //       │     │
        //     ┌─1─┐   │
        //     2   3   4
        //     T   V   U
        check(
            "((T:2,V:3):1,U:4):0;",
            LocalGeneTriplet {
                branches_lengths: [20, 30, 10, 40],
                sequence_length,
                topology: Some(T::ACB),
            },
        );

        //------------------------------------------------------------------------------------------
        // Within a larger tree.
        //
        //                   │
        //             ┌─────0───────┐
        //        ┌────1───┐         │
        //        │        │   ┌─────8──────┐
        //        │        │   │            │
        //     ┌──2──┐     │   │        ┌───10─┐
        //     │     │     │   │        │      │
        //     │   ┌─4─┐   │   │     ┌──11─┐   │
        //     │   │   │   │   │     │     │   │
        //     │   │   │   │   │   ┌─12┐   │   │
        //     3   5   6   7   9   13  14  15  16
        //     O   P   Q   X   Y   T   U   V   Z
        check(
            "(((O:3,(P:5,Q:6):4):2,X:7):1,(Y:9,(((T:13,U:14):12,V:15):11,Z:16):10):8):0;",
            LocalGeneTriplet {
                branches_lengths: [130, 140, 120, 150],
                sequence_length,
                topology: Some(T::ABC),
            },
        );

        // Discordant.
        //
        //                   │
        //             ┌─────0───────┐
        //        ┌────1───┐         │
        //        │        │   ┌─────8──────┐
        //        │        │   │            │
        //     ┌──2──┐     │   │        ┌───10─┐
        //     │     │     │   │        │      │
        //     │   ┌─4─┐   │   │     ┌──11─┐   │
        //     │   │   │   │   │     │     │   │
        //     │   │   │   │   │   ┌─12┐   │   │
        //     3   5   6   7   9   13  14  15  16
        //     O   P   Q   X   Y   U   V   T   Z
        check(
            "(((O:3,(P:5,Q:6):4):2,X:7):1,(Y:9,(((U:13,V:14):12,T:15):11,Z:16):10):8):0;",
            LocalGeneTriplet {
                branches_lengths: [130, 140, 120, 150],
                sequence_length,
                topology: Some(T::BCA),
            },
        );
    }
}
