// Parse input file into a gene trees forest.

use std::{collections::HashMap, path::Path};

use super::GenesForest;
use crate::{
    errors::Error,
    interner::Interner,
    io::read_file,
    lexer::{lexerr, Error as LexerError, Lexer},
};

impl GenesForest {
    pub fn from_file(path: &Path, interner: &mut Interner) -> Result<Self, Error> {
        Self::parse(&read_file(path)?, interner)
    }

    pub(crate) fn parse(input: &str, interner: &mut Interner) -> Result<Self, Error> {
        let mut lexer = Lexer::new(input);
        let mut trees = Vec::new();
        let mut ids = Vec::new();
        let mut index = HashMap::new();
        // Count lines to contextualize potential errors.
        let mut l = 1;
        while !lexer.trim_start().consumed() {
            let gtree = lexer.read_gene_tree(interner).map_err(with_line(l))?;
            let name = lexer.trim_start_on_line().read_block().ok_or_else(|| {
                with_line(l)(lexerr!(
                    "Unexpected end of line while reading gene tree name."
                ))
            })?;
            lexer
                .trim_start_on_line()
                .strip_char(['\n'])
                .ok_or_else(|| unexpected_data_on_line(l, &mut lexer))?;
            let symbol = interner.get_or_intern(name);
            if index.contains_key(&symbol) {
                return Err(duplicated_tree_name(l, name));
            }
            index.insert(symbol, trees.len());
            ids.push(symbol);
            trees.push(gtree);
            l += 1;
        }
        Ok(GenesForest { trees, ids, index })
    }
}

// Error conversions that require additional context.

// For use with .map_err().
fn with_line(line: usize) -> impl FnOnce(LexerError) -> Error {
    move |error| Error::GeneTree { line, error }
}

// For use from scratch.
fn duplicated_tree_name(line: usize, name: &str) -> Error {
    Error::GeneTree {
        line,
        error: LexerError::new(format!(
            "Tree name {name:?} already given earlier in the file."
        )),
    }
}
fn unexpected_data_on_line(line: usize, lexer: &mut Lexer) -> Error {
    Error::GeneTree {
        line,
        error: LexerError::new(format!(
            "Unexpected trailing data: {}.",
            lexer.read_until(['\n']).unwrap().0
        )),
    }
}
